from lazylinop import LazyLinOp
from scipy.fft import fft2, ifft2
import sys
import warnings
from warnings import warn

sys.setrecursionlimit(100000)
warnings.simplefilter(action='always')


def fft2d(in_shape: tuple):
    """
    Alias for ``dft2d(in_shape: tuple)`` function.
    """
    return dft2d(in_shape)


def dft2d(in_shape: tuple):
    """
    Returns a :class:`.LazyLinOp` ``L`` for the orthogonal
    2D Discrete-Fourier-Transform (DFT) of a 2D signal
    of shape ``in_shape = (M, N)`` (provided in flattened version).

    Shape of ``L`` is $(MN,~MN)$ with $(M,~N)=\text{in_shape}$.
    After applying the operator as ``y = L @ flatten(X)``, a 2D
    output can be obtained via ``unflatten(y, out_shape)``
    with ``out_shape = in_shape``. ``L`` is orthogonal.

    Args:
        in_shape: ``tuple``
            Shape of the 2d input array $(M,~N)$.

    Returns:
        :class:`.LazyLinOp`

    Example:
        >>> from lazylinop.signal2d import dft2d, flatten, unflatten
        >>> import numpy as np
        >>> import scipy as sp
        >>> # Check the 2d DFT
        >>> N = 32
        >>> F = dft2d((N, N))
        >>> x = np.random.rand(N, N)
        >>> ly = F @ flatten(x)
        >>> sy = sp.fft.fft2(x, norm='ortho')
        >>> np.allclose(unflatten(ly, (N, N)), sy)
        True
        >>> # Check the inverse DFT
        >>> lx_ = F.H @ ly
        >>> sx_ = sp.fft.ifft2(sy, norm='ortho')
        >>> np.allclose(lx_, flatten(sx_))
        True
        >>> # Orthogonal DFT
        >>> A = F.toarray()
        >>> H = F.H.toarray()
        >>> np.allclose(A @ H, np.eye(N ** 2, N ** 2))
        True

    .. seealso::
        - `SciPy fft2 <https://docs.scipy.org/doc/scipy/reference/
          generated/scipy.fft.fft2.html>`_,
        - :func:`lazylinop.signal.dft`.
    """

    s = in_shape[0] * in_shape[1]

    return LazyLinOp(
        shape=(s, s),
        matvec=lambda x: fft2(x.reshape(in_shape), norm='ortho').ravel(),
        rmatvec=lambda x: ifft2(x.reshape(in_shape), norm='ortho').ravel()
    )
